{*******************************************************}
{                                                       }
{     StreamSec Security Library for Embarcadero Delphi }
{     stSMIMEId10Broker Unit                            }
{                                                       }
{     For Indy 10                                       }
{                                                       }
{     Copyright (C) 2003-2013 StreamSec Handelsbolag    }
{                                                       }
{*******************************************************}
unit stSMIMEId10Broker;

interface

uses
  SysUtils, Classes,
  StreamSec.DSI.Cms,
  StreamSec.DSI.PkixCms,
  IdMessage, IdMessageParts, IdAttachment, IdAttachmentFile,
  IdAttachmentMemory, IdText, IdException;

type
  TstSMIMEIdBroker = class(TstMPCMSBroker)
  private
    class function InternalCheckIsSMIME(const aContentType: string): Boolean;
  public
    class function CheckIsMultipartSigned(aSrc: TIdMessage): Boolean;
    class function CheckIsSMIME(aSrc: TIdMessage): Boolean;
    function Decode(aSrc, aDest: TIdMessage;
                    ActionDescriptions: TStrings): Boolean; overload;
    function Encode(Actions: TContentTypeArray;
                    aSrc, aDest: TIdMessage;
                    DataType: TPKCS7DataType = pData): Boolean; overload;
  end;

procedure Register;

implementation

uses
  stBase64, stReadStrm, stOIDs, stPkixExtension, StreamSec.DSI.PkixCert;

procedure Register;
begin
  RegisterComponents('StreamSec Tools',[TstSMIMEIdBroker]);
end;

{ TstSMIMEIdBroker }

class function TstSMIMEIdBroker.CheckIsMultipartSigned(
  aSrc: TIdMessage): Boolean;
var
  lContentType: string;
  lP: Integer;
begin
  lContentType := LowerCase(aSrc.ContentType);
  lP := Pos('multipart/signed;',lContentType);
  Result := lP = 1;
  if Result then begin
    lP := Pos('protocol="application/pkcs7-signature"',lContentType);
    Result := lP > Length('multipart/signed;');
    if Result then begin
      Result := aSrc.MessageParts.Count = 2;
      if Result then begin
        lContentType := aSrc.MessageParts.Items[1].ContentType;
        Result := SameText(lContentType,'application/pkcs7-signature');
      end;
    end;
  end;
end;

class function TstSMIMEIdBroker.CheckIsSMIME(ASrc: TIdMessage): Boolean;
var
  vContentType: string;
begin
  vContentType := LowerCase(ASrc.ContentType);
  Result := InternalCheckIsSMIME(vContentType);
end;

function TstSMIMEIdBroker.Decode(ASrc, ADest: TIdMessage;
  ActionDescriptions: TStrings): Boolean;
const
  Ending: AnsiString = #13#10'.'#13#10;
var
  vContentType, vFileName, vFileExt: string;
  vSrc, vDst: TSecureMemoryStream;
  vTmpMsg: TIdMessage;
begin
  // First, test for "Content-Type: application/[x-]pkcs7-mime"
  Result := CheckIsSMIME(ASrc);
  vContentType := LowerCase(ASrc.ContentType);
  // Second, test for "Content-Disposition: attachment"
  if not Result then begin
    Result := ASrc.MessageParts.Count = 1;
    if Result then begin
      Result := ASrc.MessageParts.Items[0] is TIdAttachmentFile;
      if Result then begin
        vFileName := TIdAttachment(ASrc.MessageParts.Items[0]).FileName;
        vFileExt := LowerCase(ExtractFileExt(vFilename));
        Result := (vFileExt = '.p7m') or (vFileExt = '.p7c') or (vFileExt = '.aps');
      end;
    end else begin
      // Third, test for Content-Type: multipart/signed
      Result := CheckIsMultipartSigned(aSrc);
      if Result then begin
        Result := False;
        (*
        Result := Assigned(SignedData);
        if Result then begin
          TstMPSignedData(SignedData).DetachedSignature := True;
          vSrc := tSecureMemoryStream.Create;
          try
            vSrc.LoadFromFile(TIdAttachmentFile(aSrc.MessageParts.Items[1]).StoredPathname);
            vDst := tSecureMemoryStream.Create;
            try
              // This operation requires the raw source of this message part,
              // which is not available in Indy 10.
//              if aSrc.MessageParts.Items[0] is TIdText then
//                aSrc.SaveToStream();
              TIdAttachment(aSrc.MessageParts.Items[0]).SaveToStream(vDst);
              Result := SignedData.VerifySignedData(vSrc,vDst)
            finally
              vDst.Free;
            end;
          finally
            vSrc.Free;
          end;
          if Result then begin
            TIdAttachmentMemory(aSrc.MessageParts.Items[0]).DataStream.Position := 0;
            aDest.LoadFromStream(TIdAttachmentMemory(aSrc.MessageParts.Items[0]).DataStream);
            Exit;
          end;
        end;
        *)
      end;
    end;
  end;

  if not Result then begin
    if ASrc <> ADest then begin
      vSrc := TSecureMemoryStream.Create;
      try
        ASrc.NoEncode := True;
        ASrc.SaveToStream(vSrc);
        vSrc.Position := 0;
        ADest.LoadFromStream(vSrc);
      finally
        vSrc.Free;
      end;
    end;
    Result := True;
  end else begin
    vSrc := TSecureMemoryStream.Create;
    try
      vSrc.LoadFromFile(TIdAttachmentFile(ASrc.MessageParts.Items[0]).StoredPathname);
      vSrc.Position := 0;
      vDst := TSecureMemoryStream.Create;
      try
        Result := Decode(vSrc,vDst,ActionDescriptions);
        if Result then begin
          vDst.Seek(0,soFromEnd);
          vDst.Write(Ending[1],Length(Ending));
          vDst.Position := 0;
          vTmpMsg := ADest;
          try
            vTmpMsg.LoadFromStream(vDst);
          except
            on EIdException do
              ;
            on Exception do
              raise;
          end;
          Result := Decode(vTmpMsg,ADest,ActionDescriptions);
        end;
      finally
        vDst.Free;
      end;
    finally
      vSrc.Free;
    end;
  end;
end;

type
  TSignedDataHack = class(TstCustomMPSignedData);

function TstSMIMEIdBroker.Encode(Actions: TContentTypeArray; ASrc,
  ADest: TIdMessage; DataType: TPKCS7DataType): Boolean;
var
  Strm: TSecureMemoryStream;
  DstStrm: TOctetStringStream;
  TmpMsg: TIdMessage;
  Addrs: TStringList;
  I: Integer;
  Cert: iCertificate;
begin
  Strm := TSecureMemoryStream.Create;
  try
    if ADest.From.Address = '' then
      ADest.From.Assign(ASrc.From);
    if ADest.Recipients.Count = 0 then
      ADest.Recipients.Assign(ASrc.Recipients);
    if ADest.CCList.Count = 0 then
      ADest.CCList.Assign(ASrc.CCList);
    if ADest.BCCList.Count = 0 then
      ADest.BCCList.Assign(ASrc.BCCList);
    if ADest.Subject = '' then
      ADest.Subject := ASrc.Subject;

    TmpMsg := ADest;
    try
      if Length(Actions) = 0 then begin
        if ASrc <> ADest then begin
          ASrc.SaveToStream(Strm);
          Strm.Position := 0;
          ADest.LoadFromStream(Strm);
        end;
        Result := True;
      end else begin
        ASrc.SaveToStream(Strm);
        Strm.Position := 0;
        DstStrm := TOctetStringStream.Create('');
        try
          case Actions[0] of
            pEnvelopedData:
              begin
                TmpMsg.ContentType := 'application/pkcs7-mime;' +
                                      #9'smime-type=enveloped-data;' +
                                      #9'name="smime.p7m"';
                TmpMsg.ContentTransferEncoding := 'base64';
                TmpMsg.ContentDisposition := 'attachment; filename="smime.p7m"';
                TmpMsg.GenerateHeader;
                TmpMsg.Headers.Assign(TmpMsg.LastGeneratedHeaders);
                Result := Assigned(EnvelopedData);
                if Result then begin
                  Addrs := TStringList.Create;
                  try
                    Addrs.Sorted := True;
                    Addrs.Duplicates := dupIgnore;
                    if ADest.Sender.Address <> '' then
                      Addrs.Add(ADest.Sender.Address);
                    if ADest.From.Address <> '' then
                      Addrs.Add(ADest.From.Address);
                    for I := 0 to ADest.Recipients.Count - 1 do
                      if ADest.Recipients.Items[I].Address <> '' then
                        Addrs.Add(ADest.Recipients.Items[I].Address);
                    for I := 0 to ADest.CCList.Count - 1 do
                      if ADest.CCList.Items[I].Address <> '' then
                        Addrs.Add(ADest.CCList.Items[I].Address);
                    for I := 0 to ADest.CCList.Count - 1 do
                      if ADest.BCCList.Items[I].Address <> '' then
                        Addrs.Add(ADest.BCCList.Items[I].Address);
                    (EnvelopedData as TstMPEnvelopedData).ContentType := pData;
                    (EnvelopedData as TstMPEnvelopedData).ContentOnly := False;
                    Result := EnvelopedData.Encrypt(Addrs,Strm,DstStrm);
                  finally
                    Addrs.Free;
                  end;
                end;
              end;
            pSignedData:
              begin
                TSignedDataHack(SignedData).SigningDescription := ASrc.Subject;
                TmpMsg.ContentType := 'application/pkcs7-mime;' +
                                      #9'smime-type=signed-data;' +
                                      #9'name="smime.p7m"';
                TmpMsg.ContentTransferEncoding := 'base64';
                TmpMsg.ContentDisposition := 'attachment; filename="smime.p7m"';
                TmpMsg.GenerateHeader;
                TmpMsg.Headers.Assign(TmpMsg.LastGeneratedHeaders);
                Cert := GetCertByEMail(ADest.From.Address,rsaEncryption,[keyEncipherment,digitalSignature],True);
                if not Assigned(Cert) then begin
                  Cert := GetCertByEMail(ADest.From.Address,rsaEncryption,[digitalSignature],True);
                end;
                Result := Assigned(Cert);
                if Result then begin
                  (SignedData as TstMPSignedData).ContentType := pData;
                  (SignedData as TstMPSignedData).ContentOnly := False;
                  Result := SignedData.SignData(Cert,Strm,DstStrm);
                end;
              end;
          else
            Result := False;
          end;
          if Result then begin
            TmpMsg.Headers.Values['MIME-Version'] := '1.0';
            TmpMsg.Headers.Values['Content-Type'] := TmpMsg.ContentType;
            TmpMsg.Headers.Values['Content-Transfer-Encoding'] := TmpMsg.ContentTransferEncoding;
            TmpMsg.Headers.Values['Content-Disposition'] := TmpMsg.ContentDisposition;
            TmpMsg.Headers.Values['X-Security-Provider'] := '"StreamSec StrSecIV S/MIME Broker for Indy 10"';
            DstStrm.Position := 0;
            TmpMsg.NoEncode := True;
            TmpMsg.Body.Text := StrToMIME64(DstStrm.DataString);
            Result := Encode(Copy(Actions,1,Length(Actions)-1),TmpMsg,ADest);
          end;
        finally
          DstStrm.Free;
        end;
      end;
    finally
//
    end;
  finally
    Strm.Free;
  end;
end;

class function TstSMIMEIdBroker.InternalCheckIsSMIME(const aContentType: string): Boolean;
var
  P: Integer;
  lContentType: string;
begin
  P := Pos('application/', aContentType);
  Result := P = 1;
  if Result then begin
    P := Pos('pkcs7-mime', aContentType);
    Result := P > Length('application/');
    if Result then begin
      P := Pos('smime-type=', aContentType);
      if P > 0 then begin
        lContentType := Copy(aContentType, P + Length('smime-type='), MaxInt);
        Result := (Pos('enveloped-data', lContentType) = 1) or (Pos('signed-data', lContentType) = 1);
      end;
    end;
  end;
end;

end.
